from itertools import islice, zip_longest
import re

from log import log
from rule import Rule
from symbols import Symbol, Symbols


class Edge(Rule):
    counter = 0

    def __init__(self, predicted_from=None, predicted_by_rule=None,
                 extended_from=None, extended_using=None, tokens=None,
                 how_created=None):
        Edge.counter += 1
        self.identificator = Edge.counter

        self.how_created = how_created
        self.predicted_from = predicted_from
        self.extended_from = extended_from
        if extended_from is not None:
            self.extend_from(extended_from, extended_using, tokens)
        elif predicted_by_rule is not None:
            self.rule = predicted_by_rule
            self.left = Symbol(predicted_by_rule.left)
            self.begin = 0
            self.length = 0
            Symbols.__init__(self, predicted_by_rule, position=0)
            if predicted_from is not None:
                self.begin = predicted_from.end
                self.predict_attributes()
        self.closed = self.position == len(self)
        self.current_symbol = None if self.closed else self[self.position]

    def __eq__(self, other):
        if self.rule is not other.rule or self.position != other.position:
            return False
        elif self.left != other.left:  # tohle nedělá not __eq__!
            return False
        for symbol, other_symbol in islice(zip_longest(self, other),
                                           self.position + 1):
            if symbol != other_symbol:
                return False
        # for index, symbol in enumerate(self):
        #     if index == self.position:
        #         break
        #     elif symbol != other[index]:
        #         return False
        return True

    def __hash__(self):
        return (hash(id(self.rule)) ^ hash(self.position) ^ hash(self.begin) ^
                hash(self.length))

    def __repr__(self):
        return 'Edge(' + str(self) + ')'

    def extends(self, open_edge):
        return (
            self.begin == open_edge.end and
            self.left['phrase'] == open_edge.current_symbol.get('phrase') and
            open_edge.match_symbol(self.left))

    def extend_with(self, closed_edge, how_created):
        return Edge(extended_from=self, extended_using=closed_edge,
                    how_created=how_created)

    def extend_from(self, extended_from, extended_using, tokens):
        self.predicted_from = extended_from.predicted_from
        self.rule = extended_from.rule
        self.left = Symbol(extended_from.left)
        Symbols.__init__(self, extended_from,
                         position=extended_from.position + 1)
        updated_symbol = self[extended_from.position]
        if extended_using is not None:
            self.length += extended_using.length
            updated_symbol.update(extended_using.left)
            updated_symbol.edge = extended_using
            self.copy_attributes(extended_using.left)  # tohle u closed nebylo
        elif tokens:  # terminal rule
            self.length += 1
            updated_symbol.update(tokens[extended_from.end])
            self.copy_attributes()
            updated_symbol.begin = extended_from.end
            updated_symbol.length = 1
            updated_symbol.edge = True
        else:
            raise ArgumentError('nothing to extend {} with'.format(self))

    def predict_attributes(self):
        symbol = self.predicted_from.current_symbol
        for attr, value in self.rule.left.items():
            if value is None and attr in symbol:
                self.left[attr] = symbol[attr]
                # log.debug('copy         %s=%s → head', attr, symbol[attr])

        for attr, match_on_indexes in sorted(self.rule.attr_matches.items()):
            if -1 in match_on_indexes and attr in self.left:
                value = self.left[attr]
                for index in match_on_indexes[1:]:
                    self[index][attr] = value
                    # log.debug('copy         head → %s=%s', attr, value)

    def copy_attributes(self, update=None):
        copy_from = self.position - 1
        if update:
            current = self[copy_from]
            for attr, value in sorted(update.items()):
                if value is None:
                    log.debug('null value   %s', attr)
                    continue  # asi by se nemělo kopírovat; nejspíš nemělo
                elif attr not in current or current[attr] is None:
                    log.debug('move         %s=%s → %s', attr, value, current)
                    current[attr] = value
                elif current[attr] != value:
                    log.warning('update       %s=%s → %s', attr, value,
                                current)
                    current[attr] = value

        for attr, indexes in sorted(self.rule.attr_matches.items()):
            if copy_from not in indexes:
                continue  # odsud se ani nemá kopírovat
            elif attr not in self[copy_from]:
                log.debug('no attr      %s %s', attr, self[copy_from])
                continue  # není co kopírovat

            value = self[copy_from][attr]
            if value is None:
                log.debug('null value   %s', attr)
                continue  # asi by se nemělo kopírovat; nejspíš nemělo

            if indexes[0] == -1:  # kopírování do hlavy
                log.debug('copy         %s=%s → head', attr, value)
                self.left[attr] = value

            # kopíruje se jen na členy vpravo od právě vloženého
            for index in indexes[indexes.index(copy_from)+1:]:
                log.debug('copy         %s=%s → %s', attr, value, self[index])
                self[index][attr] = value

    # TODO: rozdělit/přesunout a zpřehlednit i v použití ve fundamentálním pr.!
    def match_symbol(self, symbol):
        if self.current_symbol is not None:
            for attr, value in self.current_symbol.items():
                # log.debug('             trying to match %s=%s in %s', attr,
                #           value, symbol)
                try:
                    if value is None:
                        pass
                    elif attr not in symbol:
                        log.debug('not present  %s in %s', attr, symbol)
                    elif symbol[attr] is None:
                        log.debug('null value   %s=%s', attr, symbol[attr])
                    elif not re.fullmatch(value, symbol[attr]):
                        # log.warning('mismatch     %s=%s should be %s in %s '
                        #             'to match %s', attr, symbol[attr], value,
                        #             symbol, self)
                        return False
                except TypeError as te:
                    log.exception(te)
                    log.debug('%s %s %s', value, attr, symbol.get(attr))
                    raise
        else:
            log.debug('no symbol    cannot match %s to %s', symbol, self)
        for attr, indexes in sorted(self.rule.attr_matches.items()):
            value = None
            for index in indexes:
                if index == -1:
                    continue
                if index < self.position:
                    if attr not in self[index]:
                        log.debug('             %s not in %s', attr,
                                    self[index])
                        continue
                    value = self[index][attr]
                elif value is not None and index == self.position:
                    if attr not in symbol:
                        log.debug('not present  %s in %s', attr, symbol)
                    elif value != symbol[attr]:
                        log.error('mismatch     %s %s != %s in %s and %s',
                                  attr, symbol[attr], value, symbol, self)
                        return False
        return True

    def match_terminal(self, token):
        # log.debug("try match    %s = %s", self.current_symbol, token)
        # TODO: tady nemůžu pracovat s "?" (optional), to by fakt chtělo
        # ukládat jinak!
        for attr, expected_value in self.current_symbol.items():
            if expected_value is None:
                continue
            elif attr not in token:
                if attr != 'phrase':
                    log.debug("no attr      %s has no %s=%s", token, attr,
                              expected_value)
                return False
            elif token[attr] is None:
                # log.debug("null attr    %s (expected: %s)", attr,
                #           expected_value)
                raise AssertionError('Attribute %s is None in %s, not %s' % (
                    attr, token, expected_value))
            try:  # elif isinstance(token[attr], str):
                if not re.fullmatch(expected_value, token[attr]):
                    # log.debug("mismatch     %s=%s (expected: %s)", attr,
                    #           token[attr], expected_value)
                    return False
            except TypeError:  # else:
                for value in token[attr]:
                    if re.fullmatch(expected_value, value):
                        break
                else:
                    log.debug("mismatch     %s=%s (expected: %s)", attr,
                              token[attr], expected_value)
                    return False

        log.debug("match        %s = %s", token, self.current_symbol)
        return True

    @property
    def begin(self):
        return self.left.begin

    @property
    def length(self):
        return self.left.length

    @property
    def end(self):
        return self.left.end

    @begin.setter
    def begin(self, begin):
        self.left.begin = begin

    @length.setter
    def length(self, length):
        self.left.length = length
