from collections import OrderedDict

from log import log
from shlex import shlex
import symbols


class XmlTag(OrderedDict):
    def __init__(self, lines=None, tag=None, opening=None, empty=None):
        super().__init__()
        self.lines = lines
        self.tag = tag
        self.opening = opening
        self.empty = empty

    @property
    def closing(self):
        return not self.opening

    def add(self, attr, value):
        existing = self.get(attr, None)
        if existing is not None:
            if isinstance(existing, str):
                self[attr] = [existing, value]
            else:
                existing.append(value)
        else:
            self[attr] = value

    def __str__(self):
        opening = '' if self.opening else '/'
        empty = '/' if self.empty else ''
        # ztratil jsem původní pořadí duplikovaných atributů
        attrs = ''.join(
            ' {}="{}"'.format(attr, val) for attr, val in self.items())
        return '<{opening}{tag}{attrs}{empty}>'.format(
            opening=opening, tag=self.tag, attrs=attrs, empty=empty)

    def __repr__(self):
        return 'XmlTag(' + symbols.escape(self) + ')'

    @classmethod
    def parse(cls, lines):
        parser = shlex(lines[1:])
        xml = cls()
        attr = None
        value = None
        while True:
            token = parser.get_token()  # ValueError("No closing quotation")
            if not token:
                break
            elif xml.opening is None:
                xml.opening = token != '/'
                if xml.opening:
                    xml.tag = token
            elif xml.tag is None:
                if token == '>':
                    break
                xml.tag = token
            elif attr is None:
                if token == '/':
                    xml.empty = True
                elif token == '>':
                    break
                elif token == '=':
                    break  # unexpected =
                else:
                    attr = token
                    value = '='  # expecting =
            elif xml.empty:
                if token == '>':
                    break
                else:
                    break  # unexpected /
            elif value == '=':
                if token == '=':
                    value = None  # expecting a value
                elif token == '/':
                    xml.add(attr, None)
                    xml.empty = True
                elif token == '>':
                    xml.add(attr, None)
                    break
                else:
                    xml.add(attr, None)  # got an empty value
                    attr = token  # immediately followed by another
            elif value is None:
                if (len(token) > 1 and token[0] in '\'"' and
                        token[0] == token[-1]):
                    token = token[1:-1]
                xml.add(attr, token)
                attr = None  # expecting an attribute
                value = None
        return xml
